"use strict";(self.webpackChunk_mlflow_mlflow=self.webpackChunk_mlflow_mlflow||[]).push([[8844],{27925:function(e,t,n){n.d(t,{w:function(){return r}});var i=n(46709);function r({ref:e,debounceTimeMs:t}){const n=(0,i.useRef)({width:-1,height:-1}),r=(0,i.useRef)(),[c,s]=(0,i.useState)(null),a=(0,i.useCallback)(e=>{if(0===e.length)return;const i=e[0].contentRect;if(-1===n.current.width)return n.current={width:i.width,height:i.height},void s(n.current);if(i.width!==n.current.width||i.height!==n.current.height){if(n.current.width=i.width,n.current.height=i.height,!t)return void s({...n.current});clearTimeout(r.current),r.current=setTimeout(()=>{s({...n.current})},t)}},[t]),l=(0,i.useRef)();return l.current||(l.current=new ResizeObserver(a)),(0,i.useEffect)(()=>{const t="function"===typeof e?e():e.current;if(t){const e=l.current;return e.observe(t),()=>e.unobserve(t)}}),c}},49633:function(e,t,n){n.d(t,{Mj:function(){return a.Mj},Ub:function(){return c},wY:function(){return s.w}});var i=n(46709),r=n(69103);function c(e){return function(e){return(0,r.useSyncExternalStore)(i.useCallback(t=>(e.addEventListener("change",t),()=>{e.removeEventListener("change",t)}),[e]),()=>e.matches)}(i.useMemo(()=>window.matchMedia(e),[e]))}var s=n(27925),a=n(53637)},78844:function(e,t,n){n.r(t);var i=n(68248),r=n(82716),c=n(27757),s=n(42747),a=n(49633),l=n(47045),u=n(19960),o=n(46709),d=n(73408);var h={name:"13z90ae",styles:"display:flex;justify-content:space-between;align-items:center;max-width:600px"};t.default=()=>{const{theme:e}=(0,c.u)(),t=(0,s.tz)(),[n,f]=(0,a.Mj)({key:l.kV,version:l.j6,initialValue:!0}),m=(0,o.useCallback)(e=>{f(e),e?u.R.start():u.R.shutdown()},[f]);return(0,d.FD)("div",{css:(0,i.AH)({padding:e.spacing.md},""),children:[(0,d.Y)(c.T.Title,{level:2,children:(0,d.Y)(s.sA,{id:"wRV8PN",defaultMessage:"Settings"})}),(0,d.FD)("div",{css:h,children:[(0,d.FD)("div",{css:(0,i.AH)({display:"flex",flexDirection:"column",marginRight:e.spacing.lg},""),children:[(0,d.Y)(c.T.Title,{level:4,children:(0,d.Y)(s.sA,{id:"hg+bcy",defaultMessage:"Enable telemetry"})}),(0,d.Y)(c.T.Text,{children:(0,d.Y)(s.sA,{id:"LXz6c5",defaultMessage:"This setting enables UI telemetry data collection. Learn more about what types of data are collected in our {documentation}.",values:{documentation:(0,d.Y)(c.T.Link,{componentId:"mlflow.settings.telemetry.documentation-link",href:"https://mlflow.org/docs/latest/community/usage-tracking.html",openInNewTab:!0,children:(0,d.Y)(s.sA,{id:"h2398a",defaultMessage:"documentation"})})}})})]}),(0,d.Y)(r.Switch,{componentId:"mlflow.settings.telemetry.toggle-switch",checked:n,onChange:m,label:" ",activeLabel:t.formatMessage({id:"+CGMk6",defaultMessage:"On"}),inactiveLabel:t.formatMessage({id:"aCzpU3",defaultMessage:"Off"}),disabledLabel:" "})]})]})}}}]);