"use strict";(self.webpackChunk_mlflow_mlflow=self.webpackChunk_mlflow_mlflow||[]).push([[4118],{60312:function(e,r,l){l.d(r,{m:function(){return i}});var a=l(82716),n=l(73408);var s={name:"13udsys",styles:"height:100%"};const i=({children:e,className:r})=>(0,n.Y)(a.PageWrapper,{css:s,className:r,children:e})},84118:function(e,r,l){l.r(r);var a=l(68248),n=l(55570),s=l(91113),i=l(884),o=l(82716),t=l(27757),d=l(40724),c=l(91105),m=l(60312),u=l(71952),g=l(85344),h=l(1664),p=l(29447),f=l(73408);var w={name:"1m5f2z8",styles:"overflow:hidden;display:flex;flex-direction:column"};r.default=(0,n.X)(s.A.mlflowServices.EXPERIMENTS,()=>{const{theme:e}=(0,t.u)(),r=(0,c.Zp)(),{form:l,isLoading:n,error:s,resetErrors:y,selectedModel:A,isFormComplete:Y,handleSubmit:b,handleCancel:v,handleNameBlur:C}=(0,u.F)({onSuccess:e=>r(p.Ay.getEndpointDetailsRoute(e.endpoint_id)),onCancel:()=>r(p.Ay.gatewayPageRoute)});return(0,f.Y)(m.m,{css:w,children:(0,f.FD)(i.Op,{...l,children:[(0,f.FD)("div",{css:(0,a.AH)({padding:e.spacing.md},""),children:[(0,f.FD)(o.Breadcrumb,{includeTrailingCaret:!0,children:[(0,f.Y)(o.Breadcrumb.Item,{children:(0,f.Y)(c.N_,{to:p.Ay.gatewayPageRoute,children:(0,f.Y)(d.A,{id:"uABFy0",defaultMessage:"AI Gateway"})})}),(0,f.Y)(o.Breadcrumb.Item,{children:(0,f.Y)(c.N_,{to:p.Ay.gatewayPageRoute,children:(0,f.Y)(d.A,{id:"eYZ/ZL",defaultMessage:"Endpoints"})})})]}),(0,f.Y)(t.T.Title,{level:2,css:(0,a.AH)({marginTop:e.spacing.sm},""),children:(0,f.Y)(d.A,{id:"UkVgwL",defaultMessage:"Create endpoint"})}),(0,f.Y)("div",{css:(0,a.AH)({marginTop:e.spacing.md,borderBottom:`1px solid ${e.colors.border}`},"")})]}),(0,f.Y)(h.V,{mode:"create",isSubmitting:n,error:s,errorMessage:(0,g.m)(s),resetErrors:y,selectedModel:A,isFormComplete:Y,onSubmit:b,onCancel:v,onNameBlur:C})]})})})}}]);