"use strict";(self.webpackChunk_mlflow_mlflow=self.webpackChunk_mlflow_mlflow||[]).push([[4205],{34205:function(e,i,t){t.r(i),t.d(i,{DiscoverNews:function(){return h},default:function(){return f}});var n=t(68248),a=t(27757),o=t(79081),r=t(40724),s=t(82716),l=t(73408);const d=[{id:"mlflow-mcp-server",title:(0,l.Y)(r.A,{id:"x/YJtF",defaultMessage:"MLflow MCP server"}),description:"Connect your coding assistants and AI applications to MLflow and automatically analyze your experiments and traces.",link:"https://mlflow.org/docs/latest/genai/mcp/index.html",componentId:"mlflow.home.news.auto_tune_llm_judge",thumbnail:{gradient:{light:"linear-gradient(135deg, #E9F2FF 0%, #F4EBFF 100%)",dark:"linear-gradient(135deg, #1E2633 0%, #2A2434 100%)"},icon:s.AssistantIcon}},{id:"optimize-prompts",title:(0,l.Y)(r.A,{id:"+erxn2",defaultMessage:"Optimize prompts"}),description:"Access the state-of-the-art prompt optimization algorithms such as MIPROv2, GEPA, through MLflow Prompt Registry.",link:"https://mlflow.org/docs/latest/genai/prompt-registry/optimize-prompts",componentId:"mlflow.home.news.optimize_prompts",thumbnail:{gradient:{light:"linear-gradient(135deg, #E8F7F2 0%, #D5E8FF 100%)",dark:"linear-gradient(135deg, #1E2B2A 0%, #1C2A3A 100%)"},icon:s.NotebookIcon}},{id:"agents-as-a-judge",title:(0,l.Y)(r.A,{id:"nRsCQz",defaultMessage:"Agents-as-a-judge"}),description:"Leverage agents as a judge to perform deep trace analysis and improve your evaluation accuracy.",link:"https://mlflow.org/docs/latest/genai/eval-monitor/scorers/llm-judge/agentic-overview/index.html",componentId:"mlflow.home.news.agents_as_a_judge",thumbnail:{gradient:{light:"linear-gradient(135deg, #FFF5E1 0%, #FFE2F2 100%)",dark:"linear-gradient(135deg, #3A2A1C 0%, #3A2230 100%)"},icon:s.RobotIcon}},{id:"dataset-tracking",title:(0,l.Y)(r.A,{id:"o8YPlv",defaultMessage:"Dataset tracking"}),description:"Track dataset lineage and versions and effectively drive the quality improvement loop.",link:"https://mlflow.org/docs/latest/genai/datasets/",componentId:"mlflow.home.news.dataset_tracking",thumbnail:{gradient:{light:"linear-gradient(135deg, #E6F3FF 0%, #E0F1F6 100%)",dark:"linear-gradient(135deg, #1F2B38 0%, #1E2A2F 100%)"},icon:s.DatabaseIcon}}];var c={name:"88wgd8",styles:"display:flex;align-items:flex-end;justify-content:flex-end;flex:1"};const g=({gradient:e,title:i,description:t,icon:r})=>{const{theme:s}=(0,a.u)();return(0,l.FD)("div",{css:(0,n.AH)({borderRadius:s.borders.borderRadiusMd,aspectRatio:"16 / 9",background:s.isDarkMode?e.dark:e.light,display:"flex",flexDirection:"column",padding:s.spacing.md,gap:s.spacing.md},""),children:[(0,l.FD)("div",{css:(0,n.AH)({display:"flex",flexDirection:"column",gap:s.spacing.xs},""),children:[(0,l.Y)(a.T.Text,{bold:!0,color:"primary",children:i}),t?(0,l.Y)(a.T.Text,{color:"secondary",children:t}):(0,l.Y)(o.S,{size:"sm"})]}),r&&(0,l.Y)("div",{css:c,children:(0,l.Y)(r,{css:(0,n.AH)({color:s.colors.white,fontSize:24},"")})})]})},m=({title:e,description:i,link:t,thumbnail:o})=>{const{theme:r}=(0,a.u)(),s={textDecoration:"none",color:r.colors.textPrimary,display:"block"},d=(0,l.Y)("div",{css:(0,n.AH)({overflow:"hidden",border:`1px solid ${r.colors.actionDefaultBorderDefault}`,borderRadius:r.borders.borderRadiusMd,background:r.colors.backgroundPrimary,padding:r.spacing.sm+r.spacing.xs,display:"flex",flexDirection:"column",gap:r.spacing.sm,boxSizing:"border-box",boxShadow:r.shadows.sm,cursor:"pointer",transition:"background 150ms ease",height:"100%",width:320,minWidth:320,"&:hover":{background:r.colors.actionDefaultBackgroundHover},"&:active":{background:r.colors.actionDefaultBackgroundPress}},""),children:(0,l.Y)(g,{gradient:o.gradient,title:e,description:i,icon:o.icon})});return(0,l.Y)("a",{href:t,target:"_blank",rel:"noopener noreferrer",css:s,children:d})};var p={name:"ti75j2",styles:"margin:0"},u={name:"1xelnrc",styles:"width:100%;min-width:0"};const h=()=>{const{theme:e}=(0,a.u)();return(0,l.FD)("section",{css:(0,n.AH)({display:"flex",flexDirection:"column",gap:e.spacing.md},""),children:[(0,l.FD)("div",{css:(0,n.AH)({display:"flex",alignItems:"center",justifyContent:"space-between",gap:e.spacing.md},""),children:[(0,l.Y)(a.T.Title,{level:3,css:p,children:(0,l.Y)(r.A,{id:"2pj5gm",defaultMessage:"Discover new features"})}),(0,l.Y)(a.T.Link,{componentId:"mlflow.home.news.view_more",href:"https://mlflow.org/blog/",openInNewTab:!0,css:(0,n.AH)({color:e.colors.textSecondary},""),children:(0,l.Y)(r.A,{id:"RgVN+O",defaultMessage:"View all"})})]}),(0,l.Y)("section",{css:u,children:(0,l.Y)("div",{css:(0,n.AH)({width:"100%",display:"flex",gap:e.spacing.sm+e.spacing.xs,flexWrap:"wrap"},""),children:d.map(e=>(0,l.Y)(m,{...e},e.id))})})]})};var f=h}}]);