"use strict";(self.webpackChunk_mlflow_mlflow=self.webpackChunk_mlflow_mlflow||[]).push([[3528],{83528:function(e,o,t){t.r(o),t.d(o,{GetStarted:function(){return g},default:function(){return p}});var i=t(68248),n=t(27757),r=t(40724),a=t(82716),s=t(73408);const l=[{id:"log-traces",icon:a.WorkflowsIcon,componentId:"mlflow.home.quick_action.log_traces",title:(0,s.Y)(r.A,{id:"wvskxE",defaultMessage:"Log traces"}),description:(0,s.Y)(r.A,{id:"sS3GBl",defaultMessage:"Trace LLM applications for debugging and monitoring."}),link:"https://mlflow.org/docs/latest/llms/tracing/index.html"},{id:"run-evaluation",icon:a.BeakerIcon,componentId:"mlflow.home.quick_action.run_evaluation",title:(0,s.Y)(r.A,{id:"GdtTc/",defaultMessage:"Run evaluation"}),description:(0,s.Y)(r.A,{id:"uXW7SK",defaultMessage:"Iterate on quality with offline evaluations and comparisons."}),link:"https://mlflow.org/docs/latest/llms/llm-evaluate/index.html"},{id:"train-models",icon:a.NotebookIcon,componentId:"mlflow.home.quick_action.train_models",title:(0,s.Y)(r.A,{id:"8xzQsr",defaultMessage:"Train models"}),description:(0,s.Y)(r.A,{id:"BZLX5c",defaultMessage:"Track experiments, parameters, and metrics throughout training."}),link:"https://mlflow.org/docs/latest/ml/tracking/quickstart/"},{id:"register-prompts",icon:a.ModelsIcon,componentId:"mlflow.home.quick_action.register_prompts",title:(0,s.Y)(r.A,{id:"Qg12xQ",defaultMessage:"Register prompts"}),description:(0,s.Y)(r.A,{id:"aqMseN",defaultMessage:"Manage prompt updates and collaborate across teams."}),link:"https://mlflow.org/docs/latest/genai/prompt-registry/"}];var c=t(86383);const d=({action:e})=>{const{theme:o}=(0,n.u)(),t={textDecoration:"none",color:o.colors.textPrimary,display:"block"},r={overflow:"hidden",border:`1px solid ${o.colors.actionDefaultBorderDefault}`,borderRadius:o.borders.borderRadiusMd,background:o.colors.backgroundPrimary,padding:o.spacing.sm+o.spacing.xs,display:"flex",gap:o.spacing.sm,width:320,minWidth:320,boxSizing:"border-box",boxShadow:o.shadows.sm,cursor:"pointer",transition:"background 150ms ease","&:hover":{background:o.colors.actionDefaultBackgroundHover},"&:active":{background:o.colors.actionDefaultBackgroundPress}},a={display:"flex",flexDirection:"column",gap:o.spacing.xs,flex:1},l={borderRadius:o.borders.borderRadiusSm,background:o.colors.actionDefaultBackgroundHover,padding:o.spacing.xs,color:o.colors.blue500,height:"min-content",display:"flex",alignItems:"center",justifyContent:"center"},d=(0,s.FD)("div",{css:r,children:[(0,s.Y)("div",{css:l,children:(0,s.Y)(e.icon,{})}),(0,s.FD)("div",{css:a,children:[(0,s.Y)("span",{role:"heading","aria-level":2,children:(0,s.Y)(n.T.Text,{bold:!0,children:e.title})}),(0,s.Y)(n.T.Text,{color:"secondary",children:e.description})]})]}),{openLogTracesDrawer:u}=(0,c.m)();return"log-traces"===e.id?(0,s.Y)("button",{type:"button",onClick:u,css:(0,i.AH)({...t,border:0,padding:0,background:"transparent",cursor:"pointer",font:"inherit",textAlign:"left"},""),children:d}):(0,s.Y)("a",{href:e.link,target:"_blank",rel:"noopener noreferrer",css:t,children:d})};var u={name:"ti75j2",styles:"margin:0"};const g=()=>{const{theme:e}=(0,n.u)();return(0,s.FD)("section",{css:(0,i.AH)({display:"flex",flexDirection:"column",gap:e.spacing.md},""),children:[(0,s.Y)(n.T.Title,{level:3,css:u,children:(0,s.Y)(r.A,{id:"O3q/U1",defaultMessage:"Get started"})}),(0,s.Y)("section",{css:(0,i.AH)({marginBottom:20,width:"100%",minWidth:0,display:"flex",gap:e.spacing.sm+e.spacing.xs,flexWrap:"wrap"},""),children:l.map(e=>(0,s.Y)(d,{action:e},e.id))})]})};var p=g}}]);