"use strict";(self.webpackChunk_mlflow_mlflow=self.webpackChunk_mlflow_mlflow||[]).push([[783],{27925:function(e,n,t){t.d(n,{w:function(){return i}});var r=t(46709);function i({ref:e,debounceTimeMs:n}){const t=(0,r.useRef)({width:-1,height:-1}),i=(0,r.useRef)(),[o,s]=(0,r.useState)(null),c=(0,r.useCallback)(e=>{if(0===e.length)return;const r=e[0].contentRect;if(-1===t.current.width)return t.current={width:r.width,height:r.height},void s(t.current);if(r.width!==t.current.width||r.height!==t.current.height){if(t.current.width=r.width,t.current.height=r.height,!n)return void s({...t.current});clearTimeout(i.current),i.current=setTimeout(()=>{s({...t.current})},n)}},[n]),a=(0,r.useRef)();return a.current||(a.current=new ResizeObserver(c)),(0,r.useEffect)(()=>{const n="function"===typeof e?e():e.current;if(n){const e=a.current;return e.observe(n),()=>e.unobserve(n)}}),o}},30783:function(e,n,t){t.r(n),t.d(n,{default:function(){return M}});var r=t(68248),i=t(46709),o=t(27757),s=t(82716),c=t(40724),a=t(60312),l=t(49708),u=t(12803),d=t(60329),h=t(26158),m=t(86383),f=t(73408);const p=i.lazy(()=>t.e(5431).then(t.bind(t,61847)));var g=()=>{const{isLogTracesDrawerOpen:e}=(0,m.m)();return e?(0,f.Y)(i.Suspense,{fallback:null,children:(0,f.Y)(p,{})}):null},w=t(47045),v=t(49633),Y=t(42747),b=t(91105);const y=()=>{const{theme:e}=(0,o.u)(),[n,t]=(0,v.Mj)({key:w.uO,version:w.Pc,initialValue:!1});return n?null:(0,f.Y)(s.Alert,{componentId:"mlflow.home.telemetry-alert",message:"Information about UI telemetry",type:"info",onClose:()=>t(!0),description:(0,f.Y)("div",{css:(0,r.AH)({display:"flex",flexDirection:"column",gap:e.spacing.sm},""),children:(0,f.Y)("span",{children:(0,f.Y)(Y.sA,{id:"qvEOHi",defaultMessage:"MLflow collects usage data to improve the product. To confirm your preferences, please visit the settings page in the navigation sidebar. To learn more about what data is collected, please visit the <documentation>documentation</documentation>.",values:{documentation:e=>(0,f.Y)(b.N_,{to:"https://mlflow.org/docs/latest/community/usage-tracking.html",target:"_blank",rel:"noopener noreferrer",children:e})}})})})})},k=i.lazy(()=>t.e(3528).then(t.bind(t,83528))),T=i.lazy(()=>t.e(4205).then(t.bind(t,34205))),x=i.lazy(()=>Promise.all([t.e(3617),t.e(547)]).then(t.bind(t,30547))),D=["home","recent-experiments"],L=()=>{const{theme:e}=(0,o.u)();return(0,f.FD)("div",{children:[(0,f.Y)(s.TitleSkeleton,{}),(0,f.Y)(s.TableSkeleton,{lines:3})]})};var M=()=>{const{theme:e}=(0,o.u)(),n=(0,h.M)(),[t,m]=(0,i.useState)(!1),{data:p,error:w,isLoading:v,refetch:Y}=(0,l.I)(D,{queryFn:()=>u.x.searchExperiments([["max_results","5"],["order_by","last_update_time DESC"]]),staleTime:3e4}),b=null===p||void 0===p?void 0:p.experiments,M=()=>m(!1);return(0,f.FD)(a.m,{css:(0,r.AH)({padding:e.spacing.md,display:"flex",flexDirection:"column",gap:e.spacing.lg,height:"min-content"},""),children:[(0,f.Y)(s.Header,{title:(0,f.Y)(c.A,{id:"sGXhV4",defaultMessage:"Welcome to MLflow"})}),(0,f.Y)(y,{}),(0,f.Y)(i.Suspense,{fallback:(0,f.Y)(L,{}),children:(0,f.Y)(k,{})}),(0,f.Y)(i.Suspense,{fallback:(0,f.Y)(L,{}),children:(0,f.Y)(x,{experiments:b,isLoading:v,error:w,onCreateExperiment:()=>m(!0),onRetry:Y})}),(0,f.Y)(i.Suspense,{fallback:(0,f.Y)(L,{}),children:(0,f.Y)(T,{})}),(0,f.Y)(d.D,{isOpen:t,onClose:M,onExperimentCreated:()=>{M(),n(),Y()}}),(0,f.Y)(g,{})]})}},49633:function(e,n,t){t.d(n,{Mj:function(){return c.Mj},Ub:function(){return o},wY:function(){return s.w}});var r=t(46709),i=t(69103);function o(e){return function(e){return(0,i.useSyncExternalStore)(r.useCallback(n=>(e.addEventListener("change",n),()=>{e.removeEventListener("change",n)}),[e]),()=>e.matches)}(r.useMemo(()=>window.matchMedia(e),[e]))}var s=t(27925),c=t(53637)},60312:function(e,n,t){t.d(n,{m:function(){return s}});var r=t(82716),i=t(73408);var o={name:"13udsys",styles:"height:100%"};const s=({children:e,className:n})=>(0,i.Y)(r.PageWrapper,{css:o,className:n,children:e})},86383:function(e,n,t){t.d(n,{m:function(){return r}});const r=(0,t(24447).vt)(e=>({isLogTracesDrawerOpen:!1,openLogTracesDrawer:()=>e({isLogTracesDrawerOpen:!0}),closeLogTracesDrawer:()=>e({isLogTracesDrawerOpen:!1})}))}}]);